@@ -29,10 +29,11 @@ import ast
 import asyncio
 import codecs
 from contextlib import contextmanager
+from enum import Enum
 import fnmatch
+import json
 import io
 from pathlib import Path
-from queue import Queue
 import sys
 from tempfile import NamedTemporaryFile
 from typing import Any, cast, Callable, Dict, Generator, List, Optional, Tuple, Union
@@ -45,17 +46,36 @@ import docutils.core
 import docutils.writers.html4css1
 from CodeChat.CodeToRest import code_to_rest_string, html_static_path  # type: ignore
 from CodeChat.CommentDelimiterInfo import SUPPORTED_GLOBS  # type: ignore
+import websockets
+
 
 # Local imports
 # -------------
-from .gen_py.CodeChat_Services.ttypes import (
-    GetResultType,
-    GetResultReturn,
-)
+# None
+
+
+# Constants
+# =========
+# The port used by a websocket connection between the CodeChat Server and the CodeChat Client.
+WEBSOCKET_PORT = 5001
+
+
+# .. _GetResultType Py:
+#
+# These must match the `constants in the client <GetResultType JS>`.
+class GetResultType(Enum):
+    html = 0
+    build = 1
+    errors = 2
+    command = 3
 
 
 # Utilities
 # =========
+def GetResultReturn(get_result_type: GetResultType, text: str):
+    return {"get_result_type": get_result_type.value, "text": text}
+
+
 # Convert a path to a URI component: make it absolute and use forward (POSIX) slashes. If the provided ``file_path`` is falsey, just return it.
 def path_to_uri(file_path: str):
     return Path(file_path).resolve().as_posix() if file_path else file_path
@@ -148,7 +168,10 @@ def _convertCodeChat(text: str, filePath: str) -> Tuple[str, str]:
         # extensions, CodeChat may not support the lexer chosen by Pygments.
         # For example, a ``.v`` file may be Verilog (supported by CodeChat)
         # or Coq (not supported). In this case, provide an error message
-        return "", "{}:ERROR: this file is not supported by CodeChat.".format(filePath)
+        return (
+            "",
+            "{}:: ERROR: this file is not supported by CodeChat.".format(filePath),
+        )
     return _convertReST(rest_string, filePath, True)
 
 
@@ -202,7 +225,7 @@ async def _convert_external(
     return stdout, stderr
 
 
-# Return False if source_file is newer than output_file; otherwise, return string with an error message.
+# _`_checkModificationTime`: Return False if source_file is newer than output_file; otherwise, return string with an error message.
 def _checkModificationTime(
     source_file: Path, base_html_file: Path, html_ext: str
 ) -> Tuple[str, str]:
@@ -223,14 +246,16 @@ def _checkModificationTime(
         else:
             return (
                 str(html_file),
-                "The file {} is older than the source file {}.".format(
-                    html_file, source_file
+                "{}:: ERROR: CodeChat renderer - source file older than the html file {}.".format(
+                    source_file, html_file
                 ),
             )
     except OSError as e:
         return (
             str(html_file),
-            "Error checking modification time of {}: {}".format(html_file, e),
+            "{}:: ERROR: CodeChat renderer - unable to check modification time of the html file {}: {}.".format(
+                source_file, html_file, e
+            ),
         )
 
 
@@ -240,7 +265,7 @@ async def _convert_external_project(
 ) -> Tuple[str, str]:
     # Run from the directory containing the project file.
     project_path = str(Path(tool_or_project_path).parent)
-    q.put(
+    await q.put(
         GetResultReturn(
             GetResultType.build,
             "Loading project file {}.\n".format(tool_or_project_path),
@@ -252,17 +277,17 @@ async def _convert_external_project(
         with open(tool_or_project_path, encoding="utf-8") as f:
             data = f.read()
     except Exception as e:
-        return "", "{}::ERROR: Unable to open. {}".format(tool_or_project_path, e)
+        return "", "{}:: ERROR: Unable to open. {}".format(tool_or_project_path, e)
 
     # Parse it and check the format
     try:
         d = ast.literal_eval(data)
     except Exception as e:
-        return "", "{}::ERROR: Unable to parse. {}".format(tool_or_project_path, e)
+        return "", "{}:: ERROR: Unable to parse. {}".format(tool_or_project_path, e)
     if not isinstance(d, dict):
         return (
             "",
-            "{}::ERROR: Unexpected type; file should contain a dict, but saw a {}".format(
+            "{}:: ERROR: Unexpected type; file should contain a dict, but saw a {}".format(
                 tool_or_project_path, type(d)
             ),
         )
@@ -270,7 +295,7 @@ async def _convert_external_project(
     if not isinstance(args, list):
         return (
             "",
-            "{}::ERROR: missing args or wrong type; saw {} (type was {}).".format(
+            "{}:: ERROR: missing args or wrong type; saw {} (type was {}).".format(
                 tool_or_project_path, args, type(args)
             ),
         )
@@ -278,7 +303,7 @@ async def _convert_external_project(
     if not isinstance(source_path, str):
         return (
             "",
-            "{}::ERROR: missing source_path or wrong type; saw {} (type was {}).".format(
+            "{}:: ERROR: missing source_path or wrong type; saw {} (type was {}).".format(
                 tool_or_project_path, source_path, type(source_path)
             ),
         )
@@ -286,7 +311,7 @@ async def _convert_external_project(
     if not isinstance(output_path, str):
         return (
             "",
-            "{}::ERROR: missing output_path or wrong type; saw {} (type was {}).".format(
+            "{}:: ERROR: missing output_path or wrong type; saw {} (type was {}).".format(
                 tool_or_project_path, output_path, type(output_path)
             ),
         )
@@ -294,7 +319,7 @@ async def _convert_external_project(
     if not isinstance(html_ext, str):
         return (
             "",
-            "{}::ERROR: wrong type for html_ext; saw {} (type was {}).".format(
+            "{}:: ERROR: wrong type for html_ext; saw {} (type was {}).".format(
                 file_path_, html_ext, type(html_ext)
             ),
         )
@@ -316,7 +341,7 @@ async def _convert_external_project(
     except Exception as e:
         return (
             "",
-            "{}::ERROR: unable to compute path relative to {}. {}".format(
+            "{}:: ERROR: unable to compute path relative to {}. {}".format(
                 file_path, source_path, e
             ),
         )
@@ -356,7 +381,9 @@ async def _run_subprocess(
     q: asyncio.Queue,
 ) -> Tuple[str, str]:
     # Explain what's going on.
-    q.put(GetResultReturn(GetResultType.build, "{} > {}\n".format(cwd, " ".join(args))))
+    await q.put(
+        GetResultReturn(GetResultType.build, "{} > {}\n".format(cwd, " ".join(args)))
+    )
 
     # Start the process.
     try:
@@ -368,7 +395,10 @@ async def _run_subprocess(
             stderr=asyncio.subprocess.PIPE,
         )
     except Exception as e:
-        return "", "external command:: ERROR:When starting. {}".format(e)
+        return (
+            "",
+            ":: ERROR: CodeChat renderer - when starting render process, {}.".format(e),
+        )
 
     # Provide a way to send stdout from the process a line at a time to the web client.
     async def stdout_streamer(stdout_stream: asyncio.StreamReader):
@@ -384,12 +414,12 @@ async def _run_subprocess(
         while True:
             ret = await stdout_stream.read(80)
             if ret:
-                q.put(GetResultReturn(GetResultType.build, decoder.decode(ret)))
+                await q.put(GetResultReturn(GetResultType.build, decoder.decode(ret)))
             else:
                 # Tell the decoder the stream is done and collect any last output.
                 s = decoder.decode(b"", True)
                 if s:
-                    q.put(GetResultReturn(GetResultType.build, s))
+                    await q.put(GetResultReturn(GetResultType.build, s))
                 break
 
     # An awaitable sequence to interact with the subprocess.
@@ -437,7 +467,7 @@ def _pass_through(text: str, file_path: str) -> Tuple[str, str]:
 
 # The "error converter" when a converter can't be found.
 def _error_converter(text: str, file_path: str) -> Tuple[str, str]:
-    return "", "{}:: ERROR: no converter found for this file.".format(file_path)
+    return "", "{}:: ERROR: No converter found for this file.".format(file_path)
 
 
 # ClientState
@@ -445,8 +475,8 @@ def _error_converter(text: str, file_path: str) -> Tuple[str, str]:
 # Store data for about each client.
 class ClientState:
     def __init__(self):
-        # A queue of messages for the client. This can be accessed by any thread.
-        self.q = Queue()
+        # A queue of messages for the client.
+        self.q = asyncio.Queue()
 
         # The remaining data in this class should only be accessed by rendering thread.
         #
@@ -567,21 +597,50 @@ async def convert_file(text: str, file_path: str, cs: ClientState) -> None:
         cs._html = html_string
 
     # Send any errors. An empty error string will clear any errors from a previous build, and should still be sent.
-    cs.q.put(GetResultReturn(GetResultType.errors, err_string))
+    await cs.q.put(GetResultReturn(GetResultType.errors, err_string))
 
     # Sending the HTML signals the end of this build.
     #
     # For Windows, make the path contain forward slashes.
     uri = path_to_uri(cs._file_path)
     # Encode this, for Windows paths which contain a colon (or unusual Linux paths).
-    cs.q.put(GetResultReturn(GetResultType.html, urllib.parse.quote(uri)))
+    await cs.q.put(GetResultReturn(GetResultType.html, urllib.parse.quote(uri)))
 
 
 # RenderManager / render thread
 # ==============================
 class RenderManager:
+    # Provide a way to perform thread-safe access of methods in this class.
+    def __getattr__(self, name: str) -> Callable:
+        if name.startswith("threadsafe_"):
+            # Strip off ``threadsafe`` and look for the function.
+            internal_func = self.__getattr__(name[11:])
+
+            # Invoke it as an async if needed.
+            async def async_wrap(*args, **kwargs):
+                return internal_func(*args, **kwargs)
+
+            # See if we need to wrap thisin an async.
+            async_func = (
+                internal_func
+                if asyncio.iscoroutinefunction(internal_func)
+                else async_wrap
+            )
+
+            # Wrap the async func in a threadsafe call.
+            def threadsafe_async(*args, **kwargs):
+                future = asyncio.run_coroutine_threadsafe(
+                    async_func(*args, **kwargs), self._loop
+                )
+                return future.result()
+
+            return threadsafe_async
+
+        # Not found. Let Python raise the exception for us.
+        return self.__getattribute__(name)
+
     # Determine if the provided id exists and is not being deleted. Return the ClientState for the id if so; otherwise, return False.
-    def _get_client_state(self, id: int) -> Union[bool, ClientState]:
+    def get_client_state(self, id: int) -> Union[bool, ClientState]:
         cs = self._client_state_dict.get(id)
         # Signal an error if this client doesn't exist or is being deleted; otherwise, return it.
         return cs if cs and not cs._is_deleting else False
@@ -597,10 +656,6 @@ class RenderManager:
 
     # Create a new client. Returns the client id on success or False on failure.
     def create_client(self) -> int:
-        future = asyncio.run_coroutine_threadsafe(self._create_client(), self._loop)
-        return future.result()
-
-    async def _create_client(self) -> int:
         if self._is_shutdown:
             return -1
         self._last_id += 1
@@ -612,12 +667,9 @@ class RenderManager:
         return id
 
     def delete_client(self, id: int) -> bool:
-        future = asyncio.run_coroutine_threadsafe(self._delete_client(id), self._loop)
-        return future.result()
-
-    async def _delete_client(self, id: int) -> bool:
-        cs = self._client_state_dict.get(id)
+        cs = self.get_client_state(id)
         if cs:
+            cs = cast(ClientState, cs)
             # Tell the worker to delete this.
             self._enqueue(id)
             cs._is_deleting = True
@@ -629,15 +681,7 @@ class RenderManager:
     def start_render(
         self, editor_text: str, file_path: str, id: int, is_dirty: bool
     ) -> bool:
-        future = asyncio.run_coroutine_threadsafe(
-            self._start_render(editor_text, file_path, id, is_dirty), self._loop
-        )
-        return future.result()
-
-    async def _start_render(
-        self, editor_text: str, file_path: str, id: int, is_dirty: bool
-    ) -> bool:
-        cs = self._get_client_state(id)
+        cs = self.get_client_state(id)
         if not cs:
             # Signal an error for an invalid client id.
             return False
@@ -655,46 +699,99 @@ class RenderManager:
         return True
 
     # Get a client's queue.
-    def get_queue(self, id: int) -> Optional[Queue]:
-        future = asyncio.run_coroutine_threadsafe(self._get_queue(id), self._loop)
-        return future.result()
-
-    async def _get_queue(self, id: int) -> Optional[Queue]:
-        cs = self._get_client_state(id)
+    def get_queue(self, id: int) -> Optional[asyncio.Queue]:
+        cs = self.get_client_state(id)
         return cast(ClientState, cs).q if cs else None
 
     # Given a URL, see if it matches with the latest render; if so, return the resulting HTML. If there's no match to the URL or the ID doesn't exist, return False. Note that the "HTML" can be None, meaning the render was stored to disk and the URL is a path to the rendered file.
     def get_render_results(self, id: int, url_path: str) -> Union[None, str, bool]:
-        future = asyncio.run_coroutine_threadsafe(
-            self._get_render_results(id, url_path), self._loop
-        )
-        return future.result()
-
-    async def _get_render_results(
-        self, id: int, url_path: str
-    ) -> Union[None, str, bool]:
-        cs = self._get_client_state(id)
+        cs = self.get_client_state(id)
         return (
             cast(ClientState, cs)._html
             if cs and path_to_uri(cast(ClientState, cs)._file_path) == url_path
             else False
         )
 
-    # Shut down the render manager.
-    def shutdown(self):
-        future = asyncio.run_coroutine_threadsafe(self._shutdown(), self._loop)
-        return future.result()
+    # Communicate with a client via a websocket.
+    async def websocket_handler(
+        self, websocket: websockets.WebSocketServerProtocol, path: str
+    ):
+        # First, read this client's ID.
+        try:
+            id = json.loads(await websocket.recv())
+        except websockets.exceptions.WebSocketException:
+            # Give up if there's a websocket error.
+            return
 
-    async def _shutdown(self):
+        # Find the queue for this client.
+        q = self.get_queue(id)
+        if not q:
+            try:
+                await websocket.send(
+                    GetResultReturn(
+                        GetResultType.command, f"error: unknown client {id}."
+                    )
+                )
+            except websockets.exceptions.WebSocketException:
+                # Ignore any errors here, since we're closing the socket anyway.
+                pass
+            return
+
+        # Send messages until shutdown.
+        while not self._is_shutdown:
+            ret = await q.get()
+            # Delete the client if this was a shutdown command.
+            if (ret["get_result_type"] == GetResultType.command) and (
+                ret["text"] == "shutdown"
+            ):
+                # Check that the queue is empty
+                if not q.empty():
+                    print(
+                        "CodeChat warning: client id {} shut down with pending commands.".format(
+                            id
+                        )
+                    )
+                # Request a client deletion.
+                self.delete_client(id)
+
+            try:
+                await websocket.send(json.dumps(ret))
+            except websockets.exceptions.WebSocketException:
+                # An error occurred -- close the websocket. The client will open another, so we can try again.
+                return
+
+    # Shut down a CodeChat client.
+    async def shutdown_client(self, id: int) -> bool:
+        q = self.get_queue(id)
+        # Fail if the ID is unknown.
+        if not q:
+            return False
+        # Send the shutdown command to the client.
+        await q.put(GetResultReturn(GetResultType.command, "shutdown"))
+        # In case the client is dead, shut down after a delay.
+        asyncio.create_task(self._delete_client_later(id))
+        # Indicate success.
+        return True
+
+    # Delete the client after a delay.
+    async def _delete_client_later(self, id: int):
+        await asyncio.sleep(1)
+        self.delete_client(id)
+
+    # Shut down the render manager.
+    async def shutdown(self):
         print("Render manager shutting down...")
         self._is_shutdown = True
         # Stop each client. This will tell the web client to shut down and also delete the render client.
-        for id, cs in self._client_state_dict.items():
-            cs.q.put(GetResultReturn(GetResultType.command, "shutdown"))
+        for id in self._client_state_dict.keys():
+            await self.shutdown_client(id)
         # A special case: there are no clients currently. Then, the code above did nothing, so shut the workers down now.
         if len(self._client_state_dict) == 0:
             for i in range(self._num_workers):
                 await self._job_q.put(None)
+        # Shut down the websocket.
+        self.websocket_server.close()
+        await self.websocket_server.wait_closed()
 
     # Start the render manager. This typically never returns.
     def run(self, *args, debug: bool = True) -> None:
@@ -720,6 +817,9 @@ class RenderManager:
         self._loop = asyncio.get_running_loop()
         self._is_shutdown = False
 
+        self.websocket_server = await websockets.serve(
+            self.websocket_handler, "127.0.0.1", WEBSOCKET_PORT
+        )
         await asyncio.gather(*[self._worker(i) for i in range(num_workers)])
 
     # Process items in the render queue.
