@@ -19,9 +19,8 @@
 # *******************************
 # |docname| - The CodeChat server
 # *******************************
-# TODO:
+# The CodeChat server receives requests from the editor/IDE, renders these, then displays them in the CodeChat client; likewise, it listens for CodeChat client requests, processes them, and forwards the results to the editor/IDE.
 #
-# - Try out `Quart <https://gitlab.com/pgjones/quart>`_.
 #
 # Imports
 # =======
@@ -30,8 +29,9 @@
 #
 # Standard library
 # ----------------
-import threading
 from pathlib import PurePosixPath
+import socket
+import threading
 import time
 from typing import Union
 import webbrowser
@@ -40,36 +40,41 @@ import webbrowser
 # -------------------
 from flask import (
     Flask,
-    request,
     Response,
     make_response,
     render_template,
     send_file,
     abort,
 )
-from thrift.protocol import TJSONProtocol  # type: ignore
-from thrift.server import TServer  # type: ignore
-from thrift.transport import TTransport  # type: ignore
+from thrift.server import TServer
+from thrift.transport import TTransport
 from thrift.transport import TSocket
 from thrift.protocol import TBinaryProtocol
-from werkzeug.serving import make_server
 
 
 # Local application imports
 # -------------------------
 from . import renderer
-from .gen_py.CodeChat_Services import EditorPlugin, CodeChatClient
+from .gen_py.CodeChat_Services import EditorPlugin
 from .gen_py.CodeChat_Services.ttypes import (
-    GetResultType,
-    GetResultReturn,
     RenderClientReturn,
     CodeChatClientLocation,
 )
 
 
+# Constants
+# =========
+# The port used for and HTTP connection from the CodeChat Client to the CodeChat Server.
+HTTP_PORT = 5000
+# .. _CodeChat service port:
+#
+# The port used for the Thrift connection between text editor/IDE extensions/plugins and the CodeChat Server. All editor/IDE plugins must use this port to access CodeChat services.
+THRIFT_PORT = 9090
+
+
 # Service provider
 # ================
-# This class implements both the EditorPlugin and CodeChat client services.
+# This class implements the EditorPlugin service.
 class CodeChatHandler:
     def __init__(self):
         self.render_manager: renderer.RenderManager
@@ -81,7 +86,7 @@ class CodeChatHandler:
                 CodeChatClientLocation._VALUES_TO_NAMES[codeChat_client_location]
             )
         )
-        id = self.render_manager.create_client()
+        id = self.render_manager.threadsafe_create_client()
         # Get the next ID.
         if id is None:
             ret = RenderClientReturn("", -1, "Duplicate id {}".format(id))
@@ -93,7 +98,7 @@ class CodeChatHandler:
             return ret
 
         # Return what's requested.
-        url = "http://127.0.0.1:5000/client?id={}".format(id)
+        url = "http://127.0.0.1:{}/client?id={}".format(HTTP_PORT, id)
         if codeChat_client_location == CodeChatClientLocation.url:
             # Just return the URL.
             ret_str = url
@@ -132,7 +137,7 @@ class CodeChatHandler:
                 path, id, is_dirty, text[:80]
             )
         )
-        if self.render_manager.start_render(text, path, id, is_dirty):
+        if self.render_manager.threadsafe_start_render(text, path, id, is_dirty):
             # Indicate success.
             print(" => (empty string)")
             return ""
@@ -141,31 +146,6 @@ class CodeChatHandler:
             print(" => {}".format(ret))
             return ret
 
-    # Pass rendered results back to the web view.
-    def get_result(self, id: int) -> GetResultReturn:
-        print("get_result(id={})".format(id))
-        q = self.render_manager.get_queue(id)
-        if not q:
-            ret = GetResultReturn(
-                GetResultType.command, "error: unknown client id {}".format(id)
-            )
-            print("  => {}".format(ret))
-            return ret
-        ret = q.get()
-        # Delete the client if this was a shutdown command.
-        if (ret.get_result_type == GetResultType.command) and (ret.text == "shutdown"):
-            # Check that the queue is empty
-            if not q.empty():
-                print(
-                    "CodeChat warning: client id {} shut down with pending commands.".format(
-                        id
-                    )
-                )
-            # Request a `client deletion`_.
-            self.render_manager.delete_client(id)
-        print("  => {}".format(ret))
-        return ret
-
     # _`Shut down an editor client`. The sequence is:
     #
     # #.    _`Client stop`: send a message to the web client, informing it of the shutdown. While the editor shouldn't make any more ``start_render`` or ``stop_client`` calls, doing so won't cause the server to misbehave.
@@ -174,13 +154,11 @@ class CodeChatHandler:
     # This method performs the first step; ``get_result`` performs the second.
     def stop_client(self, id: int) -> str:
         print("stop_client(id={})".format(id))
-        q = self.render_manager.get_queue(id)
-        if not q:
+        ok = self.render_manager.threadsafe_shutdown_client(id)
+        if not ok:
             ret = "unknown client {}.".format(id)
             print("  => {}".format(ret))
             return ret
-        # Send the shutdown command to the client.
-        q.put(GetResultReturn(GetResultType.command, "shutdown"))
         # Indicate success.
         print("  => (empty string)")
         return ""
@@ -196,31 +174,27 @@ handler = CodeChatHandler()
 # Server for the CodeChat editor plugin
 # -------------------------------------
 def editor_plugin_server() -> None:
-    transport = TSocket.TServerSocket(host="127.0.0.1", port=9090)
+    transport = TSocket.TServerSocket(host="127.0.0.1", port=THRIFT_PORT)
     tfactory = TTransport.TBufferedTransportFactory()
     pfactory = TBinaryProtocol.TBinaryProtocolFactory()
     processor = EditorPlugin.Processor(handler)
 
-    # This server spawns a thread per connection.
-    ## server = TServer.TThreadedServer(processor, transport, tfactory, pfactory)
     # Spawns up to 10 threads by default, then uses those. Mark these threads as daemon, so they will be terminated on program exit.
     server = TServer.TThreadPoolServer(
         processor, transport, tfactory, pfactory, daemon=True
     )
-    # For simplicity, we can use:
-    # server = TServer.TSimpleServer(processor, transport, tfactory, pfactory)
+    # Other options for the server:
+    #
+    # This server spawns a thread per connection.
+    ## server = TServer.TThreadedServer(processor, transport, tfactory, pfactory)  # noqa: E266
+    # An simpler server:
+    ## server = TServer.TSimpleServer(processor, transport, tfactory, pfactory)  # noqa: E266
+
     server.serve()
 
 
 # Server for the CodeChat client
 # ------------------------------
-# Create these once, globally.
-client_processor = CodeChatClient.Processor(handler)
-client_protocol = TJSONProtocol.TJSONProtocolFactory()
-client_server = TServer.TServer(
-    client_processor, None, None, None, client_protocol, client_protocol
-)
-
 client_app = Flask(
     __name__,
     # Serve CodeChat client files statically.
@@ -230,30 +204,22 @@ client_app = Flask(
     template_folder="CodeChat_Client/templates",
 )
 
+
 # The endpoint to get the HTML for the CodeChat client.
 @client_app.route("/client")
 def client_html() -> str:
-    return render_template("CodeChat_client.html")
-
-
-# The endpoint for the CodeChat client service.
-@client_app.route("/client", methods=["POST"])
-def client_service() -> Response:
-    itrans = TTransport.TMemoryBuffer(request.data)
-    otrans = TTransport.TMemoryBuffer()
-    iprot = client_server.inputProtocolFactory.getProtocol(itrans)
-    oprot = client_server.outputProtocolFactory.getProtocol(otrans)
-    client_server.processor.process(iprot, oprot)
-    return make_response(otrans.getvalue())
+    return render_template(
+        "CodeChat_client.html", WEBSOCKET_PORT=renderer.WEBSOCKET_PORT
+    )
 
 
 # The endpoint for files requested by a specific client, including rendered source files.
 @client_app.route("/client/<int:id>/<path:url_path>")
 def client_data(id: int, url_path: str) -> Union[str, Response]:
     # See if we rendered this file.
-    html = handler.render_manager.get_render_results(id, url_path)
+    html = handler.render_manager.threadsafe_get_render_results(id, url_path)
     # If we have rendered HTML, return it.
-    if html:
+    if type(html) == str:
         assert isinstance(html, str)
         response = make_response(html)
         # Don't allow the browser to cache files. See the `Flask docs <https://flask.palletsprojects.com/en/1.1.x/api/#flask.Request.cache_control>`_ and `MDN docs <https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Cache-Control#Directives>`_. TODO: allow caching of static files (but how do we know which files are static?)
@@ -283,19 +249,34 @@ def client_data(id: int, url_path: str) -> Union[str, Response]:
 # Main code
 # =========
 # Run both servers. This does not (usually) return.
-def run_servers() -> None:
+def run_servers() -> int:
+    # See if the required ports are in use, probably by another instance of this server.
+    if (
+        is_port_in_use(HTTP_PORT)
+        or is_port_in_use(renderer.WEBSOCKET_PORT)
+        or is_port_in_use(THRIFT_PORT)
+    ):
+        print(
+            f"Error: ports {HTTP_PORT}, {renderer.WEBSOCKET_PORT}, and/or {THRIFT_PORT} are already in use. Exiting."
+        )
+        return 1
+
     # Both servers block when run, so place them in a thread. Mark the Thrift server as a daemon, so it will be killed when the program shuts down.
-    editor_plugin_thread = threading.Thread(target=editor_plugin_server, daemon=True)
+    editor_plugin_thread = threading.Thread(
+        target=editor_plugin_server, name="Editor plugin server", daemon=True
+    )
     editor_plugin_thread.start()
 
     # Taken from https://stackoverflow.com/a/45017691.
-    flask_server_thread = threading.Thread(target=client_app.run, daemon=True)
+    flask_server_thread = threading.Thread(
+        target=client_app.run, name="Flask", daemon=True
+    )
     flask_server_thread.start()
 
     # Start the render loop in another thread.
     render_manager = renderer.RenderManager()
     handler.render_manager = render_manager
-    render_manager_thread = threading.Thread(target=render_manager.run)
+    render_manager_thread = threading.Thread(target=render_manager.run, name="asyncio")
     render_manager_thread.start()
 
     # Run the servers in threads until a user-requested shutdown occurs.
@@ -307,6 +288,13 @@ def run_servers() -> None:
 
     print("Shutting down...")
     # This will prevent future editor or web requests from being serviced.
-    render_manager.shutdown()
+    render_manager.threadsafe_shutdown()
     # When this is done, the Flask server and editor plugin sever can be shut down, since they're idle. Since they're daemons, they'll be shut down by exiting main.
     render_manager_thread.join()
+    return 0
+
+
+# Copied verbatim from Stack Overflow.
+def is_port_in_use(port):
+    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
+        return s.connect_ex(("localhost", port)) == 0
